@@ -39,7 +39,7 @@
 #define TPM2_PKCS11_STORE_DIR "/etc/tpm2_pkcs11"
 #endif
 
-#define DB_VERSION 4
+#define DB_VERSION 7
 
 #define goto_oom(x, l) if (!x) { LOGE("oom"); goto l; }
 #define goto_error(x, l) if (x) { goto l; }
@@ -68,6 +68,8 @@
         } \
     } while (0);
 
+#define CKR_VENDOR_SKIP (CKR_VENDOR_DEFINED | 0x01)
+
 static struct {
     sqlite3 *db;
 } global;
@@ -407,7 +409,7 @@ DEBUG_VISIBILITY int init_pobject_from_stmt(sqlite3_stmt *stmt, tpm_ctx *tpm, po
             LOGE("Expected persistent pobject config to have ESYS_TR blob");
             return SQLITE_ERROR;
         }
-        res = tpm_deserialize_handle(tpm, pobj->config.blob, &pobj->handle, NULL);
+        res = tpm_deserialize_handle(tpm, pobj->config.blob, &pobj->handle);
         if (!res) {
             /* just set a general error as rc could be success right now */
             return SQLITE_ERROR;
@@ -555,25 +557,18 @@ DEBUG_VISIBILITY WEAK void db_get_label(token *t, sqlite3_stmt *stmt, int iCol)
                         sqlite3_column_text(stmt, iCol));
 }
 
-CK_RV db_get_tokens(token **tok, size_t *len) {
+CK_RV db_get_tokens(token *tok, size_t *len) {
 
     size_t cnt = 0;
 
-    token *tmp = calloc(MAX_TOKEN_CNT, sizeof(token));
-    if (!tmp) {
-        LOGE("oom");
-        return CKR_HOST_MEMORY;
-    }
-
     const char *sql =
             "SELECT * FROM tokens";
 
-    sqlite3_stmt *stmt;
+    sqlite3_stmt *stmt = NULL;
     int rc = sqlite3_prepare_v2(global.db, sql, -1, &stmt, NULL);
     if (rc != SQLITE_OK) {
-        free(tmp);
         LOGE("Cannot prepare tobject query: %s\n", sqlite3_errmsg(global.db));
-        return CKR_GENERAL_ERROR;
+        goto error;
     }
 
     size_t row = 0;
@@ -584,7 +579,7 @@ CK_RV db_get_tokens(token **tok, size_t *len) {
             goto error;
         }
 
-        token *t = &tmp[row++];
+        token *t = &tok[row++];
         int col_count = sqlite3_data_count(stmt);
 
         int i;
@@ -651,15 +646,17 @@ CK_RV db_get_tokens(token **tok, size_t *len) {
         cnt++;
     }
 
-    *tok = tmp;
     *len = cnt;
     sqlite3_finalize(stmt);
 
     return CKR_OK;
 
 error:
-    token_free_list(tmp, cnt);
-    sqlite3_finalize(stmt);
+    token_free_list(&tok, &cnt);
+    *len = 0;
+    if (stmt) {
+        sqlite3_finalize(stmt);
+    }
     return CKR_GENERAL_ERROR;
 }
 
@@ -961,7 +958,7 @@ error:
     return rv;
 }
 
-CK_RV db_update_tobject_attrs(unsigned id, attr_list *attrs) {
+CK_RV _db_update_tobject_attrs(sqlite3 *db, unsigned id, attr_list *attrs) {
     assert(attrs);
 
     CK_RV rv = CKR_GENERAL_ERROR;
@@ -978,9 +975,9 @@ CK_RV db_update_tobject_attrs(unsigned id, attr_list *attrs) {
           "UPDATE tobjects SET"
             " attrs=?"      // index: 1 type: TEXT (JSON)
             " WHERE id=?;";  // Index 2 type: int
-    int rc = sqlite3_prepare_v2(global.db, sql, -1, &stmt, NULL);
+    int rc = sqlite3_prepare_v2(db, sql, -1, &stmt, NULL);
     if (rc != SQLITE_OK) {
-        LOGE("%s", sqlite3_errmsg(global.db));
+        LOGE("%s", sqlite3_errmsg(db));
         goto error;
     }
 
@@ -1004,6 +1001,12 @@ error:
     return rv;
 }
 
+CK_RV db_update_tobject_attrs(unsigned id, attr_list *attrs) {
+    assert(attrs);
+
+    return _db_update_tobject_attrs(global.db, id,  attrs);
+}
+
 CK_RV db_add_token(token *tok) {
     assert(tok);
 
@@ -1691,6 +1694,443 @@ error:
     return rv;
 }
 
+/*
+ * A bug in the python code added CBC_PAD twice and missed CTR mode, filter out the CBC_PAD
+ * and CTR mode and add them both back in properly, ie one of each in CKA_ALLOWED_MECHANISMS.
+ */
+static CK_RV handle_AES_add_cbc_ctr_modes(tobject *tobj) {
+
+    CK_OBJECT_CLASS cka_class = attr_list_get_CKA_CLASS(tobj->attrs, CK_OBJECT_CLASS_BAD);
+    if (cka_class != CKO_SECRET_KEY) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_KEY_TYPE cka_key_type = attr_list_get_CKA_KEY_TYPE(tobj->attrs, CKA_KEY_TYPE_BAD);
+    if (cka_key_type != CKK_AES) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(tobj->attrs, CKA_ALLOWED_MECHANISMS);
+
+    CK_ULONG mech_num = a ? (a->ulValueLen/sizeof(CK_MECHANISM_TYPE)) : 0;
+    /* plus two: one for CBC_PAD one for CTR modes */
+    safe_adde(mech_num, 2);
+
+    CK_MECHANISM_TYPE *new_mechs = calloc(mech_num, sizeof(CK_MECHANISM_TYPE));
+    if (!new_mechs) {
+        return CKR_HOST_MEMORY;
+    }
+
+    /* copy the old mechanism into the list and add CBC_PAD and CTR */
+    CK_ULONG i;
+    CK_ULONG pos = 0;
+    for (i=0; i < mech_num; i++) {
+        assert(a && a->pValue && a->ulValueLen);
+        CK_MECHANISM_TYPE old_mech = ((CK_MECHANISM_TYPE_PTR)(a->pValue))[i];
+        if (old_mech == CKM_AES_CBC_PAD ||
+                old_mech == CKM_AES_CTR) {
+            /*
+             * don't add the ones were adding, just to make the add below
+             * unconditional. This is an unlikely case as someone would have
+             * had to twiddle their object config by hand.
+             */
+            continue;
+        }
+
+        new_mechs[pos++] = old_mech;
+    }
+
+
+    /* append the missing mechanisms */
+    new_mechs[pos++] = CKM_AES_CBC_PAD;
+    new_mechs[pos++] = CKM_AES_CTR;
+
+    CK_ULONG total_bytes = 0;
+    safe_mul(total_bytes, pos, sizeof(CK_MECHANISM_TYPE));
+
+    CK_ATTRIBUTE updated_attr = {
+        .type = CKA_ALLOWED_MECHANISMS,
+        .ulValueLen = total_bytes,
+        .pValue = new_mechs
+    };
+
+    /*
+     * if the object had CKA_ALLOWED_MECHANISMS we UPDATE the entry,
+     * else we just add it to the list
+     */
+    CK_RV rv = a ? attr_list_update_entry(tobj->attrs, &updated_attr) :
+            attr_list_append_entry(&tobj->attrs, &updated_attr);
+    if (rv != CKR_OK) {
+        SAFE_FREE(new_mechs);
+        return rv;
+    }
+
+    return CKR_OK;
+}
+
+static CK_RV dbup_handler_from_4_to_5(sqlite3 *updb) {
+
+    /*
+     * Between version 3 and 4 of the DB the following changes need to be made:
+     *
+     * Table tobjects:
+     *
+     * The YAML attributes need to include CKM_AES_CBC_PAD and CKM_AES_CTR in the CKM_ALLOWED_MECHANISMS list.
+     */
+
+    CK_RV rv = CKR_GENERAL_ERROR;
+    sqlite3_stmt *stmt = NULL;
+
+    int rc = sqlite3_prepare_v2(updb, "SELECT * from tobjects", -1, &stmt, 0);
+    if (rc != SQLITE_OK) {
+        LOGE("Failed to fetch data: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    rc = sqlite3_step(stmt);
+    if (rc == SQLITE_DONE) {
+        goto out;
+    } else if (rc != SQLITE_ROW) {
+        LOGE("Failed to step: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    while (rc == SQLITE_ROW) {
+
+        CK_MECHANISM_TYPE *new_mechs = NULL;
+
+        tobject *tobj = db_tobject_new(stmt);
+        if (!tobj) {
+            LOGE("Could not process tobjects for upgrade");
+            goto error;
+        }
+
+        /*
+         * Due to a bug in Python side of adding AES objects, we need to run the
+         * CKA_ALLOWED_MECHANISM updates to add modes CTR and CBC on both the 4->5
+         * and 5->6 updates in the C code. Since the logic is the same for AES,
+         * we just call the routine. It was handled properly in the C code in 4->5
+         * but the Python code added CTR twice and missed CBC, so the 5->6 update
+         * gets that one.
+         */
+        rv = handle_AES_add_cbc_ctr_modes(tobj);
+        if (rv == CKR_VENDOR_SKIP) {
+            goto next;
+        } else if (rv != CKR_OK) {
+            tobject_free(tobj);
+            goto error;
+        }
+
+        /* persist the changes in the update db */
+        rv = _db_update_tobject_attrs(updb, tobj->id, tobj->attrs);
+        if (rv != CKR_OK) {
+            SAFE_FREE(new_mechs);
+            tobject_free(tobj);
+            goto error;
+        }
+
+next:
+        tobject_free(tobj);
+        SAFE_FREE(new_mechs);
+
+        rc = sqlite3_step(stmt);
+        if (rc != SQLITE_ROW && rc != SQLITE_DONE) {
+            LOGE("Failed to fetch data: %s\n", sqlite3_errmsg(updb));
+            goto error;
+        }
+    }
+
+out:
+    rv = CKR_OK;
+
+error:
+    sqlite3_finalize(stmt);
+    return rv;
+}
+
+static CK_RV handle_ECDSA_5_to_6(tobject *tobj) {
+
+    CK_OBJECT_CLASS cka_class = attr_list_get_CKA_CLASS(tobj->attrs, CK_OBJECT_CLASS_BAD);
+    if (cka_class != CKO_PRIVATE_KEY) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_KEY_TYPE cka_key_type = attr_list_get_CKA_KEY_TYPE(tobj->attrs, CKA_KEY_TYPE_BAD);
+    if (cka_key_type != CKK_EC) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(tobj->attrs, CKA_ALLOWED_MECHANISMS);
+
+    CK_ULONG mech_num = a ? (a->ulValueLen/sizeof(CK_MECHANISM_TYPE)) : 0;
+    /* plus three: SHA256 SHA384 and SHA512 */
+    safe_adde(mech_num, 3);
+
+    CK_MECHANISM_TYPE *new_mechs = calloc(mech_num, sizeof(CK_MECHANISM_TYPE));
+    if (!new_mechs) {
+        return CKR_HOST_MEMORY;
+    }
+
+    /* copy the old mechanism into the list and add CBC_PAD and CTR */
+    CK_ULONG i;
+    CK_ULONG pos = 0;
+    for (i=0; i < mech_num; i++) {
+        assert(a && a->pValue && a->ulValueLen);
+        CK_MECHANISM_TYPE old_mech = ((CK_MECHANISM_TYPE_PTR)(a->pValue))[i];
+        if (old_mech == CKM_ECDSA_SHA256 ||
+                old_mech == CKM_ECDSA_SHA384 ||
+                old_mech == CKM_ECDSA_SHA512) {
+            /*
+             * don't add the ones were adding, just to make the add below
+             * unconditional. This is an unlikely case as someone would have
+             * had to twiddle their object config by hand.
+             */
+            continue;
+        }
+
+        new_mechs[pos++] = old_mech;
+    }
+
+    /* append the missing mechanisms */
+    new_mechs[pos++] = CKM_ECDSA_SHA256;
+    new_mechs[pos++] = CKM_ECDSA_SHA384;
+    new_mechs[pos++] = CKM_ECDSA_SHA512;
+
+    CK_ULONG total_bytes = 0;
+    safe_mul(total_bytes, pos, sizeof(CK_MECHANISM_TYPE));
+
+    CK_ATTRIBUTE updated_attr = {
+        .type = CKA_ALLOWED_MECHANISMS,
+        .ulValueLen = total_bytes,
+        .pValue = new_mechs
+    };
+
+    /*
+     * if the object had CKA_ALLOWED_MECHANISMS we UPDATE the entry,
+     * else we just add it to the list
+     */
+    CK_RV rv = a ? attr_list_update_entry(tobj->attrs, &updated_attr) :
+            attr_list_append_entry(&tobj->attrs, &updated_attr);
+    if (rv != CKR_OK) {
+        SAFE_FREE(new_mechs);
+        return rv;
+    }
+
+    return CKR_OK;
+}
+
+static CK_RV dbup_handler_from_5_to_6(sqlite3 *updb) {
+
+    /*
+     * Between version 3 and 4 of the DB the following changes need to be made:
+     *
+     * Table tobjects:
+     *
+     * The YAML attributes need to include CKM_AES_CBC_PAD and CKM_AES_CTR in the CKM_ALLOWED_MECHANISMS list.
+     */
+
+    CK_RV rv = CKR_GENERAL_ERROR;
+    sqlite3_stmt *stmt = NULL;
+
+    int rc = sqlite3_prepare_v2(updb, "SELECT * from tobjects", -1, &stmt, 0);
+    if (rc != SQLITE_OK) {
+        LOGE("Failed to fetch data: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    rc = sqlite3_step(stmt);
+    if (rc == SQLITE_DONE) {
+        goto out;
+    } else if (rc != SQLITE_ROW) {
+        LOGE("Failed to step: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    while (rc == SQLITE_ROW) {
+
+
+        tobject *tobj = db_tobject_new(stmt);
+        if (!tobj) {
+            LOGE("Could not process tobjects for upgrade");
+            goto error;
+        }
+
+        rv = handle_ECDSA_5_to_6(tobj);
+        if (rv == CKR_OK) {
+            goto handled;
+        } else if (rv != CKR_VENDOR_SKIP) {
+            /* actual error */
+            tobject_free(tobj);
+            goto error;
+        }
+
+        rv = handle_AES_add_cbc_ctr_modes(tobj);
+        if (rv == CKR_OK) {
+            goto handled;
+        } else if (rv == CKR_VENDOR_SKIP) {
+            goto next;
+        } else {
+            /* actual error */
+            tobject_free(tobj);
+            goto error;
+        }
+
+handled:
+        /* persist the changes in the update db */
+        rv = _db_update_tobject_attrs(updb, tobj->id, tobj->attrs);
+        if (rv != CKR_OK) {
+            tobject_free(tobj);
+            goto error;
+        }
+
+next:
+        tobject_free(tobj);
+
+        rc = sqlite3_step(stmt);
+        if (rc != SQLITE_ROW && rc != SQLITE_DONE) {
+            LOGE("Failed to fetch data: %s\n", sqlite3_errmsg(updb));
+            goto error;
+        }
+    }
+
+out:
+    rv = CKR_OK;
+
+error:
+    sqlite3_finalize(stmt);
+    return rv;
+}
+
+static CK_RV handle_EC_AES_drop_0_allowed_mechs(tobject *tobj) {
+
+    CK_OBJECT_CLASS cka_class = attr_list_get_CKA_CLASS(tobj->attrs, CK_OBJECT_CLASS_BAD);
+    if (cka_class != CKO_PRIVATE_KEY && cka_class != CKO_SECRET_KEY) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_KEY_TYPE cka_key_type = attr_list_get_CKA_KEY_TYPE(tobj->attrs, CKA_KEY_TYPE_BAD);
+    if (cka_key_type != CKK_EC && cka_key_type != CKK_AES) {
+        return CKR_VENDOR_SKIP;
+    }
+
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(tobj->attrs, CKA_ALLOWED_MECHANISMS);
+
+    /*
+     * The new array will be at most the same size as the old array as theirs nothing
+     * to filer out.
+     */
+    CK_ULONG mech_num = a ? (a->ulValueLen/sizeof(CK_MECHANISM_TYPE)) : 0;
+
+    CK_MECHANISM_TYPE *new_mechs = calloc(mech_num, sizeof(CK_MECHANISM_TYPE));
+    if (!new_mechs) {
+        return CKR_HOST_MEMORY;
+    }
+
+    /* copy the old mechanism into the list and drop 0 mechanisms */
+    CK_ULONG i;
+    CK_ULONG pos = 0;
+    for (i=0; i < mech_num; i++) {
+        assert(a && a->pValue && a->ulValueLen);
+        CK_MECHANISM_TYPE old_mech = ((CK_MECHANISM_TYPE_PTR)(a->pValue))[i];
+        if (old_mech == 0) {
+            /*
+             * don't add the ones were adding, just to make the add below
+             * unconditional. This is an unlikely case as someone would have
+             * had to twiddle their object config by hand.
+             */
+            continue;
+        }
+
+        new_mechs[pos++] = old_mech;
+    }
+
+    CK_ULONG total_bytes = 0;
+    safe_mul(total_bytes, pos, sizeof(CK_MECHANISM_TYPE));
+
+    CK_ATTRIBUTE updated_attr = {
+        .type = CKA_ALLOWED_MECHANISMS,
+        .ulValueLen = total_bytes,
+        .pValue = new_mechs
+    };
+
+    /*
+     * if the object had CKA_ALLOWED_MECHANISMS we UPDATE the entry,
+     * else we just add it to the list
+     */
+    CK_RV rv = a ? attr_list_update_entry(tobj->attrs, &updated_attr) :
+            attr_list_append_entry(&tobj->attrs, &updated_attr);
+    if (rv != CKR_OK) {
+        SAFE_FREE(new_mechs);
+        return rv;
+    }
+
+    return CKR_OK;
+}
+
+static CK_RV dbup_handler_from_6_to_7(sqlite3 *updb) {
+
+    /*
+     * Between version 3 and 4 of the DB the following changes need to be made:
+     *
+     * Table tobjects:
+     *
+     * The YAML attributes need to include CKM_AES_CBC_PAD and CKM_AES_CTR in the CKM_ALLOWED_MECHANISMS list.
+     */
+
+    CK_RV rv = CKR_GENERAL_ERROR;
+    sqlite3_stmt *stmt = NULL;
+
+    int rc = sqlite3_prepare_v2(updb, "SELECT * from tobjects", -1, &stmt, 0);
+    if (rc != SQLITE_OK) {
+        LOGE("Failed to fetch data: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    rc = sqlite3_step(stmt);
+    if (rc == SQLITE_DONE) {
+        goto out;
+    } else if (rc != SQLITE_ROW) {
+        LOGE("Failed to step: %s", sqlite3_errmsg(updb));
+        goto error;
+    }
+
+    while (rc == SQLITE_ROW) {
+
+        tobject *tobj = db_tobject_new(stmt);
+        if (!tobj) {
+            LOGE("Could not process tobjects for upgrade");
+            goto error;
+        }
+
+        rv = handle_EC_AES_drop_0_allowed_mechs(tobj);
+        if (rv != CKR_VENDOR_SKIP) {
+            /* actual error */
+            tobject_free(tobj);
+            goto error;
+        }
+
+        /* persist the changes in the update db */
+        rv = _db_update_tobject_attrs(updb, tobj->id, tobj->attrs);
+        tobject_free(tobj);
+        if (rv != CKR_OK) {
+            goto error;
+        }
+
+        rc = sqlite3_step(stmt);
+        if (rc != SQLITE_ROW && rc != SQLITE_DONE) {
+            LOGE("Failed to fetch data: %s\n", sqlite3_errmsg(updb));
+            goto error;
+        }
+    }
+
+out:
+    rv = CKR_OK;
+
+error:
+    sqlite3_finalize(stmt);
+    return rv;
+}
+
 static CK_RV db_backup(sqlite3 *db, const char *dbpath, sqlite3 **updb, char **copypath) {
 
     CK_RV rv = CKR_GENERAL_ERROR;
@@ -1763,6 +2203,9 @@ static CK_RV db_update(sqlite3 **xdb, const char *dbpath, unsigned old_version,
             dbup_handler_from_1_to_2,
             dbup_handler_from_2_to_3,
             dbup_handler_from_3_to_4,
+            dbup_handler_from_4_to_5,
+            dbup_handler_from_5_to_6,
+            dbup_handler_from_6_to_7
     };
 
     /*
@@ -1793,7 +2236,7 @@ static CK_RV db_update(sqlite3 **xdb, const char *dbpath, unsigned old_version,
      */
     size_t i;
     for(i=old_version; i < ARRAY_LEN(updaters) && i < new_version; i++) {
-        CK_RV rv = updaters[i](dbbak);
+        rv = updaters[i](dbbak);
         if (rv != CKR_OK) {
             LOGE("Running updater index %zu failed", i);
             goto out;
@@ -1814,7 +2257,7 @@ static CK_RV db_update(sqlite3 **xdb, const char *dbpath, unsigned old_version,
      * Swap the current db with the backup db, by:
      * 1. closing them
      * 2. renaming the current db as .old
-     * 3. renaming the bakup db as the old db
+     * 3. renaming the backup db as the old db
      * 4. unlinking the .old db
      * 5. opening the new db
      */
