@@ -35,6 +35,7 @@ enum mechanism_flags {
     mf_ecc           = 1 << 10,
     mf_aes           = 1 << 11,
     mf_force_synthetic = 1 << 12,
+    mf_hmac          = 1 << 13,
 };
 
 typedef struct mdetail_entry mdetail_entry;
@@ -106,6 +107,7 @@ static CK_RV rsa_pss_hash_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list
 static CK_RV ecc_keygen_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs);
 static CK_RV ecdsa_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs);
 static CK_RV hash_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs);
+static CK_RV hmac_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs);
 static CK_RV rsa_pkcs_synthesizer(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs,
         CK_BYTE_PTR inbuf, CK_ULONG inlen,
         CK_BYTE_PTR outbuf, CK_ULONG_PTR outlen);
@@ -168,7 +170,10 @@ static const mdetail_entry _g_mechs_templ[] = {
 
     { .type = CKM_ECDSA,           .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_tpm_opdata = tpm_ec_ecdsa_get_opdata },
 
-    { .type = CKM_ECDSA_SHA1,      .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_halg = sha1_get_halg, .get_digester = sha1_get_digester, .get_tpm_opdata = tpm_ec_ecdsa_sha1_get_opdata },
+    { .type = CKM_ECDSA_SHA1,      .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_halg = sha1_get_halg, .get_digester = sha1_get_digester,     .get_tpm_opdata = tpm_ec_ecdsa_sha1_get_opdata },
+    { .type = CKM_ECDSA_SHA256,    .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_halg = sha256_get_halg, .get_digester = sha256_get_digester, .get_tpm_opdata = tpm_ec_ecdsa_sha256_get_opdata },
+    { .type = CKM_ECDSA_SHA384,    .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_halg = sha384_get_halg, .get_digester = sha384_get_digester, .get_tpm_opdata = tpm_ec_ecdsa_sha384_get_opdata },
+    { .type = CKM_ECDSA_SHA512,    .flags = mf_sign|mf_verify|mf_ecc, .validator = ecdsa_validator, .get_halg = sha512_get_halg, .get_digester = sha512_get_digester, .get_tpm_opdata = tpm_ec_ecdsa_sha512_get_opdata },
 
     /* AES */
     { .type = CKM_AES_KEY_GEN, .flags = mf_is_keygen|mf_aes },
@@ -184,6 +189,12 @@ static const mdetail_entry _g_mechs_templ[] = {
     { .type = CKM_SHA256, .flags = mf_is_digester|mf_aes, .validator = hash_validator, .get_digester = sha256_get_digester },
     { .type = CKM_SHA384, .flags = mf_is_digester|mf_aes, .validator = hash_validator, .get_digester = sha384_get_digester },
     { .type = CKM_SHA512, .flags = mf_is_digester|mf_aes, .validator = hash_validator, .get_digester = sha512_get_digester },
+
+    /* hmac */
+    { .type = CKM_SHA_1_HMAC,  .flags = mf_sign|mf_verify|mf_hmac, .validator = hmac_validator, .get_tpm_opdata = tpm_hmac_sha1_get_opdata   },
+    { .type = CKM_SHA256_HMAC, .flags = mf_sign|mf_verify|mf_hmac, .validator = hmac_validator, .get_tpm_opdata = tpm_hmac_sha256_get_opdata },
+    { .type = CKM_SHA384_HMAC, .flags = mf_sign|mf_verify|mf_hmac, .validator = hmac_validator, .get_tpm_opdata = tpm_hmac_sha384_get_opdata },
+    { .type = CKM_SHA512_HMAC, .flags = mf_sign|mf_verify|mf_hmac, .validator = hmac_validator, .get_tpm_opdata = tpm_hmac_sha512_get_opdata },
 };
 
 const rsa_detail _g_rsa_keysizes_templ [] = {
@@ -443,10 +454,38 @@ CK_RV hash_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs) {
     }
 
     /* all known hashing digests are supported in software */
-
     return CKR_OK;
 }
 
+CK_RV hmac_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs) {
+    UNUSED(attrs);
+    UNUSED(m);
+
+    /* The non general HMAC routines don't take params */
+    if (mech->pParameter || mech->ulParameterLen) {
+        return CKR_MECHANISM_PARAM_INVALID;
+    }
+
+    CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(attrs, CKA_ALLOWED_MECHANISMS);
+    if (!a) {
+        LOGE("Expected CKA_ALLOWED_MECHANISMS");
+        return CKR_GENERAL_ERROR;
+    }
+
+    /* Is the mechanism in the list of allowed mechs? */
+    bool supported = false;
+    CK_ULONG i;
+    for (i=0; i < _L(a); i++) {
+        CK_MECHANISM_TYPE t = _P(a)[i];
+        if (t == mech->mechanism) {
+            supported = true;
+            break;
+        }
+    }
+
+    return supported ? CKR_OK : CKR_MECHANISM_INVALID;
+}
+
 CK_RV rsa_pkcs_validator(mdetail *m, CK_MECHANISM_PTR mech, attr_list *attrs) {
     UNUSED(m);
 
@@ -1335,6 +1374,23 @@ CK_RV mech_is_hashing_needed(mdetail *m,
     return CKR_OK;
 }
 
+CK_RV mech_is_HMAC(mdetail *m, CK_MECHANISM_PTR mech, bool *is_hmac) {
+
+    check_pointer(m);
+    check_pointer(mech);
+    check_pointer(is_hmac);
+
+    mdetail_entry *d = mlookup(m, mech->mechanism);
+    if (!d) {
+        LOGE("Mechanism not supported, got: 0x%lx", mech->mechanism);
+        return CKR_MECHANISM_INVALID;
+    }
+
+    *is_hmac = !!(d->flags & mf_hmac);
+
+    return CKR_OK;
+}
+
 CK_RV mech_is_hashing_knowledge_needed(mdetail *m,
     CK_MECHANISM_PTR mech,
     bool *is_hashing_knowledge_needed) {
@@ -1526,7 +1582,32 @@ CK_RV mech_get_info(mdetail *m, tpm_ctx *tctx,
         return get_ecc_mechinfo(tctx, info);
     }
 
+    if (d->flags & mf_hmac) {
+        /*
+         * peering into TPM internals, the code seems to make the key the length
+         * of the hash alg, however, I could nto find this in the spec.
+         */
+        size_t len = utils_get_halg_size(mech_type);
+        info->ulMinKeySize = info->ulMaxKeySize = len;
+        return CKR_OK;
+    }
+
     LOGE("Unknown mechanism, got: 0x%lx", mech_type);
 
     return CKR_MECHANISM_INVALID;
 }
+
+CK_RV mech_is_ecc(mdetail *m, CK_MECHANISM_TYPE mech_type, bool *is_ecc) {
+
+    *is_ecc = false;
+
+    mdetail_entry *d = mlookup(m, mech_type);
+    if (!d) {
+        LOGE("Mechanism not supported, got: 0x%lx", mech_type);
+        return CKR_MECHANISM_INVALID;
+    }
+
+    *is_ecc = !!(d->flags & mf_ecc);
+
+    return CKR_OK;
+}
