@@ -83,9 +83,10 @@ CK_RV object_mech_is_supported(tobject *tobj, CK_MECHANISM_PTR mech) {
     return CKR_MECHANISM_INVALID;
 }
 
-CK_RV tobject_get_max_buf_size(tobject *tobj, size_t *maxsize) {
+CK_RV tobject_get_min_buf_size(tobject *tobj, CK_MECHANISM_PTR mech, size_t *maxsize) {
 
     assert(tobj);
+    assert(mech);
 
     CK_ATTRIBUTE_PTR a = attr_get_attribute_by_type(tobj->attrs, CKA_KEY_TYPE);
     if (!a) {
@@ -176,6 +177,49 @@ CK_RV tobject_get_max_buf_size(tobject *tobj, size_t *maxsize) {
         return CKR_OK;
     }
 
+    if (key_type == CKK_SHA_1_HMAC) {
+        *maxsize = 20;
+        return CKR_OK;
+    }
+
+    if (key_type == CKK_SHA256_HMAC) {
+        *maxsize = 32;
+        return CKR_OK;
+    }
+
+    if (key_type == CKK_SHA384_HMAC) {
+        *maxsize = 48;
+        return CKR_OK;
+    }
+
+    /* Best guess on buffer size for CKK_GENERIC_SECRET */
+    if (key_type == CKK_SHA512_HMAC) {
+        *maxsize = 64;
+        return CKR_OK;
+    }
+
+    if (key_type == CKK_GENERIC_SECRET) {
+
+        switch (mech->mechanism) {
+            case CKM_SHA_1_HMAC:
+                *maxsize = 20;
+                return CKR_OK;
+            case CKM_SHA256_HMAC:
+                *maxsize = 32;
+                return CKR_OK;
+            case CKM_SHA384_HMAC:
+                *maxsize = 48;
+                return CKR_OK;
+            case CKM_SHA512_HMAC:
+                *maxsize = 64;
+                return CKR_OK;
+            default:
+                LOGE("Cannot handle CKK_GENERIC_SECRET Mechanism: 0x%lx",
+                        mech->mechanism);
+                return CKR_MECHANISM_INVALID;
+        }
+    }
+
     LOGE("Unknown signing key type, got: 0x%lx", key_type);
 
     return CKR_GENERAL_ERROR;
@@ -393,8 +437,8 @@ CK_RV object_find(session_ctx *ctx, CK_OBJECT_HANDLE *object, CK_ULONG max_objec
         // Get the current object, and grab it's id for the object handle
         CK_OBJECT_HANDLE handle = opdata->cur->tobj_handle;
 
-        // filter out CKA_PRIVATE set to CK_TRUE if not logged in
-        if (opdata->cur->cka_private && !token_is_user_logged_in(tok)) {
+        // filter out CKA_PRIVATE set to CK_TRUE if not logged in and PIN is needed
+        if (opdata->cur->cka_private && !token_is_user_logged_in(tok) && !tok->config.empty_user_pin) {
             opdata->cur = opdata->cur->next;
             continue;
         }
@@ -816,7 +860,7 @@ attr_list *tobject_get_attrs(tobject *tobj) {
     return tobj->attrs;
 }
 
-CK_RV tobject_user_increment(tobject *tobj) {
+CK_RV _tobject_user_increment(tobject *tobj, const char *filename, int lineno) {
 
     if (tobj->active == UINT_MAX) {
        LOGE("tobject active at max count, cannot issue. id: %u", tobj->id);
@@ -824,11 +868,12 @@ CK_RV tobject_user_increment(tobject *tobj) {
     }
 
     tobj->active++;
+    _LOGV(filename, lineno, "Incremented tobject id: %u, value: %u", tobj->id, tobj->active);
 
     return CKR_OK;
 }
 
-CK_RV tobject_user_decrement(tobject *tobj) {
+CK_RV _tobject_user_decrement(tobject *tobj, const char *filename, int lineno) {
 
     if (!tobj->active) {
         LOGE("Returning a non-active tobject id: %u", tobj->id);
@@ -836,6 +881,7 @@ CK_RV tobject_user_decrement(tobject *tobj) {
     }
 
     tobj->active--;
+    _LOGV(filename, lineno, "Decremented tobject id: %u, value: %u", tobj->id, tobj->active);
 
     return CKR_OK;
 }
