@@ -267,7 +267,7 @@ bool handle_attr_event(yaml_event_t *event,
         free(state->s->seqbuf);
         state->s->seqbuf = NULL;
         if (!res) {
-            LOGE("Cannot add seqence to attr list: 0x%lx", state->s->key);
+            LOGE("Cannot add sequence to attr list: 0x%lx", state->s->key);
             return res;
         }
         return pop_handler(state);
@@ -436,6 +436,9 @@ bool handle_token_config_event(yaml_event_t *e,
             } else if(!strcmp(state->key, "pss-sigs-good")) {
                 config->pss_sigs_good = !strcmp((const char *)e->data.scalar.value, "true")
                         ? pss_config_state_good : pss_config_state_bad;
+            } else if(!strcmp(state->key, "empty-user-pin")) {
+                config->empty_user_pin = !strcmp((const char *)e->data.scalar.value, "true")
+                        ? true : false;
             } else {
                 LOGE("Unknown key, got: \"%s\"\n",
                         state->key);
