@@ -199,9 +199,20 @@ static CK_RV common_init(operation op, session_ctx *ctx, CK_MECHANISM_PTR mechan
         }
     }
 
-    /* TPM is only used on sign operations, not verify */
+    /* TPM is only used on sign operations, not verify. This is because
+     * with asymmetric key pairs we can use the public key for verification in SW
+     * without needing to involve the TPM. This with HMAC we need to use the TPM
+     * verify since the key is only resident in the TPM.
+     */
+    bool is_hmac = false;
+    rv = mech_is_HMAC(tok->mdtl, mechanism, &is_hmac);
+    if (rv != CKR_OK) {
+        LOGE("Could not determine if algorithm is HMAC or not");
+        return rv;
+    }
+
     tpm_op_data *tpm_opdata = NULL;
-    if (op == operation_sign) {
+    if (op == operation_sign || is_hmac) {
 
         rv = update_pss_sig_state(tok, tobj);
         if (rv != CKR_OK) {
@@ -232,7 +243,8 @@ static CK_RV common_init(operation op, session_ctx *ctx, CK_MECHANISM_PTR mechan
         return CKR_HOST_MEMORY;
     }
 
-    if (op != operation_sign) {
+    /* Use SW Verify if it's an asymmetric key with pkey set */
+    if (op != operation_sign && opdata->pkey) {
         opdata->crypto_opdata->use_sw = true;
         rv = sw_encrypt_data_init(tok->mdtl,
                 mechanism, tobj, &opdata->crypto_opdata->cryptopdata.sw_enc_data);
@@ -323,19 +335,19 @@ CK_RV sign_final_ex(session_ctx *ctx, CK_BYTE_PTR signature, CK_ULONG_PTR signat
 
     twist digest_buf = NULL;
 
-    size_t tmp_len = 0;
-    rv = tobject_get_max_buf_size(tobj, &tmp_len);
+    size_t expected_sig_len = 0;
+    rv = tobject_get_min_buf_size(tobj, &opdata->mech, &expected_sig_len);
     if (rv != CKR_OK) {
         return rv;
     }
 
     if (!signature) {
-        *signature_len = tmp_len;
+        *signature_len = expected_sig_len;
         goto out;
     }
 
-    if (*signature_len < tmp_len) {
-        *signature_len = tmp_len;
+    if (*signature_len < expected_sig_len) {
+        *signature_len = expected_sig_len;
         rv = CKR_BUFFER_TOO_SMALL;
         goto out;
     }
@@ -355,6 +367,7 @@ CK_RV sign_final_ex(session_ctx *ctx, CK_BYTE_PTR signature, CK_ULONG_PTR signat
             LOGE("Hash algorithm cannot have 0 size");
             return CKR_GENERAL_ERROR;
         }
+
         digest_buf = twist_calloc(hash_len);
         if (!digest_buf) {
             LOGE("oom");
@@ -376,6 +389,18 @@ CK_RV sign_final_ex(session_ctx *ctx, CK_BYTE_PTR signature, CK_ULONG_PTR signat
     CK_ULONG syn_buf_len = sizeof(syn_buf);
     CK_ULONG digest_buf_len = twist_len(digest_buf);
 
+    bool is_ecc = false;
+    rv = mech_is_ecc(tok->mdtl, opdata->mech.mechanism, &is_ecc);
+    if (rv != CKR_OK) {
+        LOGE("COuld not determine if mechanism is ECC: %lu", rv);
+        return rv;
+    }
+
+    if (is_ecc && (digest_buf_len > expected_sig_len)) {
+        LOGV("Truncating hash for EC Signature from %lu to %lu", digest_buf_len, expected_sig_len);
+        digest_buf_len = expected_sig_len;
+    }
+
     rv = mech_synthesize(
             tok->mdtl,
             &opdata->mech, tobj->attrs,
@@ -538,29 +563,31 @@ CK_RV verify_final (session_ctx *ctx, CK_BYTE_PTR signature, CK_ULONG signature_
     tobject *tobj = session_ctx_opdata_get_tobject(ctx);
     assert(tobj);
 
-    // TODO mode to buffer size
-    CK_BYTE hash[1024];
-    CK_ULONG hash_len = sizeof(hash);
+    CK_ULONG data_len = 0;
+    CK_BYTE_PTR data = NULL;
 
+    /* buffer has to stay in scope for assignment to data pointer */
+    CK_BYTE _buffer[1024];
     if (opdata->do_hash) {
-        rv = digest_final_op(ctx, opdata->digest_opdata, hash, &hash_len);
+        // TODO mode to buffer size
+        CK_ULONG _buffer_len = sizeof(_buffer);
+        rv = digest_final_op(ctx, opdata->digest_opdata, _buffer, &_buffer_len);
         if (rv != CKR_OK) {
             goto out;
         }
+        data_len = _buffer_len;
+        data = _buffer;
     } else {
-        size_t datalen = twist_len(opdata->buffer);
-        if (datalen > hash_len) {
-            LOGE("Internal buffer too small, got: %zu expected less than %zu",
-                    datalen, hash_len);
-            rv = CKR_GENERAL_ERROR;
-            goto out;
-        }
-        hash_len = datalen;
-        memcpy(hash, opdata->buffer, datalen);
+        data_len = twist_len(opdata->buffer);
+        data = (const CK_BYTE_PTR)opdata->buffer;
     }
 
-    rv = ssl_util_sig_verify(opdata->pkey, opdata->padding, opdata->md,
-            hash, hash_len, signature, signature_len);
+    if (opdata->pkey) {
+        rv = ssl_util_sig_verify(opdata->pkey, opdata->padding, opdata->md,
+            data, data_len, signature, signature_len);
+    } else {
+        rv = tpm_verify(opdata->crypto_opdata->cryptopdata.tpm_opdata, data, data_len, signature, signature_len);
+    }
 
 out:
     assert(tobj);
