@@ -354,24 +354,32 @@ size_t utils_get_halg_size(CK_MECHANISM_TYPE mttype) {
             /* falls-thru */
         case CKM_SHA1_RSA_PKCS_PSS:
             /* falls-thru */
+        case CKM_SHA_1_HMAC:
+            /* falls-thru */
         case CKM_SHA_1:
             return 20;
         case CKM_SHA256_RSA_PKCS:
             /* falls-thru */
         case CKM_SHA256_RSA_PKCS_PSS:
             /* falls-thru */
+        case CKM_SHA256_HMAC:
+            /* falls-thru */
         case CKM_SHA256:
             return 32;
         case CKM_SHA384_RSA_PKCS:
             /* falls-thru */
         case CKM_SHA384_RSA_PKCS_PSS:
             /* falls-thru */
+        case CKM_SHA384_HMAC:
+            /* falls-thru */
         case CKM_SHA384:
             return 48;
         case CKM_SHA512_RSA_PKCS:
             /* falls-thru */
         case CKM_SHA512_RSA_PKCS_PSS:
             /* falls-thru */
+        case CKM_SHA512_HMAC:
+            /* falls-thru */
         case CKM_SHA512:
             return 64;
     }
@@ -454,3 +462,114 @@ CK_RV ec_params_to_nid(CK_ATTRIBUTE_PTR ecparams, int *nid) {
 
     return CKR_OK;
 }
+
+CK_RV apply_pkcs7_pad(const CK_BYTE_PTR in, CK_ULONG inlen,
+        CK_BYTE_PTR out, CK_ULONG_PTR outlen) {
+
+    size_t total_bytes = (inlen + 16) - (inlen % 16);
+
+    if (*outlen < total_bytes) {
+        return CKR_BUFFER_TOO_SMALL;
+    }
+
+    int pad_value = total_bytes - inlen;
+
+    memcpy(out, in, inlen);
+    memset(&out[inlen], pad_value, pad_value);
+
+    *outlen = total_bytes;
+
+    return CKR_OK;
+}
+
+CK_RV remove_pkcs7_pad(CK_BYTE_PTR in, CK_ULONG inlen,
+        CK_BYTE_PTR out, CK_ULONG_PTR outlen) {
+
+    if (inlen % 16) {
+        LOGE("AES_CBC_PAD data should be block sized, got: %lu", inlen);
+        return CKR_ENCRYPTED_DATA_LEN_RANGE;
+    }
+
+    CK_BYTE pad_value = in[inlen - 1];
+    if (pad_value < 1 || pad_value > 16) {
+        LOGE("Nonesensical pad value, got: %u, expected 1-16", pad_value);
+        return CKR_ENCRYPTED_DATA_INVALID;
+    }
+
+    CK_ULONG new_size = inlen - pad_value;
+
+    if (new_size > *outlen) {
+        return CKR_BUFFER_TOO_SMALL;
+    }
+
+    *outlen = new_size;
+    memcpy(out, in, new_size);
+
+    return CKR_OK;
+}
+
+void parse_lib_version(const char *userbuf, CK_BYTE *major, CK_BYTE *minor) {
+
+    if (!userbuf || !userbuf[0]) {
+        *major = *minor = 0;
+        return;
+    }
+
+    bool is_release = strchr(userbuf, '-') == NULL;
+    if (!is_release) {
+        *major = *minor = 0;
+        return;
+    }
+
+    /* don't modify the caller */
+    char *buf = strdup(userbuf);
+    if (!buf) {
+        LOGE("oom");
+        *major = *minor = 0;
+        return;
+    }
+
+    char *minor_str = "0";
+    const char *major_str = &buf[0];
+
+
+    char *split = strchr(buf, '.');
+    if (split) {
+        split[0] = '\0';
+        minor_str = &split[1];
+        split = strchr(minor_str, '.');
+        if (split) {
+            split[0]='\0';
+        }
+    }
+
+    if (!major_str[0] || !minor_str[0]) {
+        *major = *minor = 0;
+        goto out;
+    }
+
+    char *endptr = NULL;
+    unsigned long val;
+    errno = 0;
+    val = strtoul(major_str, &endptr, 10);
+    if (errno != 0 || endptr[0] || val > UINT8_MAX) {
+        LOGW("Could not strtoul(%s): %s", major_str, strerror(errno));
+        *major = *minor = 0;
+        goto out;
+    }
+
+    *major = val;
+
+    endptr = NULL;
+    val = strtoul(minor_str, &endptr, 10);
+    if (errno != 0 || endptr[0] || val > UINT8_MAX) {
+        LOGW("Could not strtoul(%s): %s", minor_str, strerror(errno));
+        *major = *minor = 0;
+        goto out;
+    }
+
+    *minor = val;
+
+out:
+    free(buf);
+}
