@@ -1,11 +1,22 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 
-#include <openssl/evp.h>
-#include <openssl/rsa.h>
 #include <openssl/bn.h>
 #include <openssl/err.h>
+#include <openssl/evp.h>
+#include <openssl/rsa.h>
+#include <openssl/sha.h>
 
+#include "largebin.h"
 #include "test.h"
+/*
+* This HMAC key is static in the fixtures folder.
+*/
+static const unsigned char hmac_key[] = {
+   0x30, 0x33, 0x33, 0x36, 0x61, 0x61, 0x37, 0x39,
+   0x34, 0x35, 0x61, 0x33, 0x63, 0x61, 0x64, 0x65,
+   0x63, 0x33, 0x63, 0x62, 0x64, 0x63, 0x36, 0x65,
+   0x37, 0x39, 0x30, 0x34, 0x33, 0x62, 0x35, 0x62
+};
 
 struct test_info {
     CK_SESSION_HANDLE handle;
@@ -415,7 +426,7 @@ static void test_sign_verify_CKM_ECDSA(void **state) {
     /* The signature comes back as DER encoded R + S parts of the signature.
      * R + S is 2 times the curve size in bytes (so 64 for P256) but we're not
      * returning that, but the DER encoded format that tools expect.
-     * Since the length of DER encoding is dependend on the encoded value
+     * Since the length of DER encoding is dependent on the encoded value
      * (e.g. leading zero if negative), the output size is not stable.
      * Thus calling C_Sign for size must return the maximum length of the DER
      * encoded value, which is (2+1+keylength) * 2 + 2. So for P256 = 72
@@ -495,6 +506,174 @@ static void test_sign_verify_CKM_ECDSA_SHA1(void **state) {
     assert_int_equal(rv, CKR_OK);
 }
 
+static void test_sign_verify_CKM_ECDSA_SHA256(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_OBJECT_CLASS key_class = CKO_PRIVATE_KEY;
+    CK_KEY_TYPE key_type = CKK_EC;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class) },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    user_login(session);
+
+    /* Find an EC key */
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_ECDSA_SHA256 };
+
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_BYTE ckm_ecdsa_sha256_sig[4096];
+    CK_ULONG ckm_ecdsa_sha256_siglen = 0;
+
+    /* Call C_Sign for Size */
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            NULL, &ckm_ecdsa_sha256_siglen);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_not_equal(ckm_ecdsa_sha256_siglen, 0);
+
+    CK_ULONG tmp_len = ckm_ecdsa_sha256_siglen;
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha256_sig, &ckm_ecdsa_sha256_siglen);
+    assert_int_equal(rv, CKR_OK);
+    /* actual size must not be larger than previously indicated */
+    assert_in_range(ckm_ecdsa_sha256_siglen, 1, tmp_len);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha256_sig, ckm_ecdsa_sha256_siglen);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void test_sign_verify_CKM_ECDSA_SHA384(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_OBJECT_CLASS key_class = CKO_PRIVATE_KEY;
+    CK_KEY_TYPE key_type = CKK_EC;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class) },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    user_login(session);
+
+    /* Find an EC key */
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_ECDSA_SHA256 };
+
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_BYTE ckm_ecdsa_sha384_sig[4096];
+    CK_ULONG ckm_ecdsa_sha384_siglen = 0;
+
+    /* Call C_Sign for Size */
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            NULL, &ckm_ecdsa_sha384_siglen);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_not_equal(ckm_ecdsa_sha384_siglen, 0);
+
+    CK_ULONG tmp_len = ckm_ecdsa_sha384_siglen;
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha384_sig, &ckm_ecdsa_sha384_siglen);
+    assert_int_equal(rv, CKR_OK);
+    /* actual size must not be larger than previously indicated */
+    assert_in_range(ckm_ecdsa_sha384_siglen, 1, tmp_len);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha384_sig, ckm_ecdsa_sha384_siglen);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void test_sign_verify_CKM_ECDSA_SHA512(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    CK_OBJECT_CLASS key_class = CKO_PRIVATE_KEY;
+    CK_KEY_TYPE key_type = CKK_EC;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class) },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    user_login(session);
+
+    /* Find an EC key */
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_ECDSA_SHA256 };
+
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_BYTE ckm_ecdsa_sha512_sig[4096];
+    CK_ULONG ckm_ecdsa_sha512_siglen = 0;
+
+    /* Call C_Sign for Size */
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            NULL, &ckm_ecdsa_sha512_siglen);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_not_equal(ckm_ecdsa_sha512_siglen, 0);
+
+    CK_ULONG tmp_len = ckm_ecdsa_sha512_siglen;
+    rv = C_Sign(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha512_sig, &ckm_ecdsa_sha512_siglen);
+    assert_int_equal(rv, CKR_OK);
+    /* actual size must not be larger than previously indicated */
+    assert_in_range(ckm_ecdsa_sha512_siglen, 1, tmp_len);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR ) _data, sizeof(_data),
+            ckm_ecdsa_sha512_sig, ckm_ecdsa_sha512_siglen);
+    assert_int_equal(rv, CKR_OK);
+}
+
 static void test_pss(CK_SESSION_HANDLE session,
         CK_OBJECT_HANDLE key[2], CK_MECHANISM_TYPE halg) {
 
@@ -1127,11 +1306,426 @@ static void test_cert_no_good(void **state) {
     assert_int_equal(rv, CKR_OK);
 }
 
+static void test_sign_verify_CKM_SHA256_HMAC(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_BYTE label[] = "hmac0";
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_SHA256_HMAC;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+        { CKA_LABEL, &label, sizeof(label) - 1 },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA256_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const CK_BYTE_PTR msg = (typeof(msg))"Hello World This is my message to HMAC";
+    CK_ULONG msg_len = strlen((const char *)msg);
+    CK_BYTE sig[32] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, msg_len, sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void ossl_verify_hmac_sig(const unsigned char *hmac_key, size_t hmac_key_len,
+        const EVP_MD *md,
+        const CK_BYTE_PTR msg, CK_ULONG msg_len,
+        CK_BYTE_PTR sig, CK_ULONG sig_len) {
+    EVP_PKEY *ekey = EVP_PKEY_new_mac_key(EVP_PKEY_HMAC, NULL, hmac_key,
+                                   hmac_key_len);
+    assert_non_null(ekey);
+
+    EVP_MD_CTX *mdctx = EVP_MD_CTX_new();
+    assert_non_null(mdctx);
+
+    int rc = EVP_DigestSignInit(mdctx, NULL, md, NULL, ekey);
+    assert_int_equal(rc, 1);
+
+    rc = EVP_DigestSignUpdate(mdctx, msg, msg_len);
+    assert_int_equal(rc, 1);
+
+    unsigned char sig2[32] = { 0 };
+    size_t sig2_len = sizeof(sig2);
+    rc = EVP_DigestSignFinal(mdctx, sig2, &sig2_len);
+    assert_int_equal(rc, 1);
+
+    EVP_MD_CTX_free(mdctx);
+    EVP_PKEY_free(ekey);
+
+    assert_int_equal(sig2_len, sig_len);
+    assert_memory_equal(sig2, sig, sig2_len);
+}
+
+static void test_sign_verify_CKM_SHA256_HMAC_imported(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_BYTE label[] = "imported_hmac_key";
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+        { CKA_LABEL, &label, sizeof(label) - 1 },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA256_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const CK_BYTE_PTR msg = (typeof(msg))"Hello World This is my message to HMAC";
+    CK_ULONG msg_len = strlen((const char *)msg);
+    CK_BYTE sig[32] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+
+    /*
+     * verify with OpenSSL. We do this here instead of in a script because sign via command line tools
+     * like pkcs11-tool is not supported at the moment. However support was added here:
+     *   - https://github.com/OpenSC/OpenSC/pull/2385
+     */
+
+    ossl_verify_hmac_sig(hmac_key, sizeof(hmac_key),
+        EVP_sha256(),
+        msg, msg_len,
+        sig, sig_len);
+}
+
+static void test_sign_verify_CKM_SHA256_HMAC_large(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_BYTE label[] = "hmac0";
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_SHA256_HMAC;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+        { CKA_LABEL, &label, sizeof(label) - 1 },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA256_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const CK_BYTE_PTR msg = _large_rand_bin;
+    CK_ULONG msg_len = sizeof(_large_rand_bin);
+    CK_BYTE sig[32] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, msg_len, sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+
+    /* Try 5.2 style returns */
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_SignUpdate(session, (CK_BYTE_PTR)msg, msg_len);
+    assert_int_equal(rv, CKR_OK);
+
+    sig_len = 0;
+    rv = C_SignFinal(session, NULL, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_SignFinal(session, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    sig_len = 0;
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, NULL, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, msg_len, sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void test_sign_verify_CKM_SHA256_HMAC_imported_large(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_BYTE label[] = "imported_hmac_key";
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+        { CKA_LABEL, &label, sizeof(label) - 1 },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA256_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const CK_BYTE_PTR msg = _large_rand_bin;
+    CK_ULONG msg_len = sizeof(_large_rand_bin);
+    CK_BYTE sig[32] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, msg_len, sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 32);
+
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+
+    ossl_verify_hmac_sig(hmac_key, sizeof(hmac_key),
+        EVP_sha256(),
+        msg, msg_len,
+        sig, sig_len);
+}
+
+static void test_sign_verify_CKM_SHA_1_HMAC(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA_1_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const char *msg = "Hello World This is my message to HMAC";
+    CK_BYTE sig[20] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 20);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void test_sign_verify_CKM_SHA384_HMAC(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA384_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const char *msg = "Hello World This is my message to HMAC";
+    CK_BYTE sig[48] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 48);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+}
+
+static void test_sign_verify_CKM_SHA512_HMAC(void **state) {
+
+    test_info *ti = test_info_from_state(state);
+    CK_SESSION_HANDLE session = ti->handle;
+
+    user_login(session);
+
+    CK_OBJECT_CLASS key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_GENERIC_SECRET;
+    CK_ATTRIBUTE tmpl[] = {
+        { CKA_CLASS, &key_class, sizeof(key_class)  },
+        { CKA_KEY_TYPE, &key_type, sizeof(key_type) },
+    };
+
+       /* FIND A generic key for HMAC */
+    CK_RV rv = C_FindObjectsInit(session, tmpl, ARRAY_LEN(tmpl));
+    assert_int_equal(rv, CKR_OK);
+
+    CK_ULONG count;
+    CK_OBJECT_HANDLE objhandles[1];
+    rv = C_FindObjects(session, objhandles, ARRAY_LEN(objhandles), &count);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(count, 1);
+
+    rv = C_FindObjectsFinal(session);
+    assert_int_equal(rv, CKR_OK);
+
+    CK_MECHANISM mech = { .mechanism = CKM_SHA512_HMAC };
+    rv = C_SignInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    const char *msg = "Hello World This is my message to HMAC";
+    CK_BYTE sig[64] = { 0 };
+    CK_ULONG sig_len = sizeof(sig);
+    rv = C_Sign(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, &sig_len);
+    assert_int_equal(rv, CKR_OK);
+    assert_int_equal(sig_len, 64);
+
+    rv = C_VerifyInit(session, &mech, objhandles[0]);
+    assert_int_equal(rv, CKR_OK);
+
+    rv = C_Verify(session, (CK_BYTE_PTR)msg, (CK_ULONG)strlen(msg), sig, sig_len);
+    assert_int_equal(rv, CKR_OK);
+    rv = C_Logout(session);
+    assert_int_equal(rv, CKR_OK);
+}
+
 int main() {
 
     const struct CMUnitTest tests[] = {
         cmocka_unit_test_setup_teardown(test_sign_verify_CKM_RSA_PKCS_PSS,
-                test_setup, test_teardown),
+            test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_sign_verify_context_specific_good,
             test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_sign_verify_public,
@@ -1152,10 +1746,30 @@ int main() {
             test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_sign_verify_CKM_ECDSA_SHA1,
             test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_ECDSA_SHA256,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_ECDSA_SHA384,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_ECDSA_SHA512,
+            test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_sign_verify_CKM_ECDSA,
             test_setup, test_teardown),
         cmocka_unit_test_setup_teardown(test_cert_no_good,
             test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA_1_HMAC,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA384_HMAC,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA256_HMAC,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA256_HMAC_imported,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA256_HMAC_large,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA256_HMAC_imported_large,
+            test_setup, test_teardown),
+        cmocka_unit_test_setup_teardown(test_sign_verify_CKM_SHA512_HMAC,
+            test_setup, test_teardown),
     };
 
     return cmocka_run_group_tests(tests, group_setup, group_teardown);
