@@ -31,7 +31,8 @@ setup(
 
     # Dependencies got here aka install_requires=['tensorflow']
     install_requires=[
-        'cryptography',
+        'bcrypt',
+        'cryptography>=3.0',
         'pyyaml',
         'pyasn1',
         'pyasn1_modules',
