@@ -54,12 +54,12 @@ class NewKeyCommandBase(Command):
             '--hierarchy-auth',
             help='The hierarchyauth, required for transient pobjects.\n',
             default='')
-        pinopts = group_parser.add_mutually_exclusive_group(required=True)
+        pinopts = group_parser.add_mutually_exclusive_group()
         pinopts.add_argument('--sopin', help='The Administrator pin.\n'),
         pinopts.add_argument('--userpin', help='The User pin.\n'),
 
     # Implemented by derived class
-    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):
         raise NotImplementedError('Implement: new_key')
 
     def new_key_init(self, label, sopin, userpin, hierarchyauth, pobj, sealobjects, tpm2, d):
@@ -96,7 +96,8 @@ class NewKeyCommandBase(Command):
 
     @staticmethod
     def new_key_save(alg, keylabel, tid, label, privblob, pubblob,
-                     tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=None, extra_pubattrs=None):
+                     tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=None, extra_pubattrs=None,
+                     override_keylen=None):
         token = db.gettoken(label)
 
         #
@@ -129,7 +130,8 @@ class NewKeyCommandBase(Command):
         if initial_pubattrs and extra_pubattrs:
             initial_pubattrs.update(extra_pubattrs)
 
-        objects = PKCS11ObjectFactory(y, tpm2, encobjauth, initial_pubattrs, initial_privattrs, tpm_pub=pubblob, tpm_priv=privblob)
+        objects = PKCS11ObjectFactory(y, tpm2, encobjauth, initial_pubattrs, initial_privattrs,
+                      tpm_pub=pubblob, tpm_priv=privblob, override_keylen=override_keylen)
 
         # Store private to database
         db.addtertiary(token['id'], objects['private'])
@@ -167,6 +169,7 @@ class NewKeyCommandBase(Command):
                 key_label = args['key_label']
                 tid = args['id']
                 hierarchyauth = args['hierarchy_auth']
+                passin = args['passin'] if 'passin' in args else None
 
                 privkey = None
                 try:
@@ -178,6 +181,14 @@ class NewKeyCommandBase(Command):
                 pobjectid = token['pid']
                 pobj = db.getprimary(pobjectid)
 
+                if userpin is None and sopin is None:
+                    # Use empty PIN if the token has an empty user PIN
+                    token_config = yaml.safe_load(io.StringIO(token['config']))
+                    if token_config.get('empty-user-pin'):
+                        userpin = ''
+                    else:
+                        sys.exit('error: at least one of the arguments --sopin --userpin is required')
+
                 sealobjects = db.getsealobject(token['id'])
 
                 encobjauth, objauth = self.new_key_init(
@@ -185,15 +196,18 @@ class NewKeyCommandBase(Command):
                     pobj, sealobjects, tpm2, d)
 
                 tertiarypriv, tertiarypub, tertiarypubdata = self.new_key_create(
-                    pobj, objauth, hierarchyauth, tpm2, alg, privkey, d)
+                    pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d)
 
                 # handle options that can add additional attributes
                 always_auth = args['attr_always_authenticate']
                 priv_attrs = {CKA_ALWAYS_AUTHENTICATE : always_auth}
 
+                override_keylen = getattr(self, '_override_keylen', None)
+
                 return NewKeyCommandBase.new_key_save(
                     alg, key_label, tid, label, tertiarypriv, tertiarypub,
-                    tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=priv_attrs)
+                    tertiarypubdata, encobjauth, db, tpm2, extra_privattrs=priv_attrs,
+                    override_keylen=override_keylen)
 
 
 @commandlet("import")
@@ -213,16 +227,33 @@ class ImportCommand(NewKeyCommandBase):
         group_parser.add_argument(
             '--algorithm',
             help='The type of the key.\n',
-            choices=['rsa', 'ecc'],
-            required=True)
+            choices=['rsa', 'ecc',
+                     'hmac', # Generic hmac, actuall imported as keyedhash scheme null with proper attrs.
+                     'hmac:sha1', 'hmac:sha256', 'hmac:sha384', 'hmac:sha512' ],
+            required=False)
+        group_parser.add_argument(
+            '--passin',
+            help='Password of the input private key file like OpenSSL (example: "pass:secret").\n',
+            required=False)
 
     # Imports a new key
-    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):
 
         pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
 
+        # A plain HMAC key is just a keyedhash object with userwithauth and sign set.
+        # Seal objects don't have sign (ie default tpm2_import -G keyedhash behavior.
+        objattrs=None
+        if alg == 'hmac':
+            objattrs = 'userwithauth|sign|decrypt'
+            alg = 'keyedhash'
+
         tertiarypriv, tertiarypub, tertiarypubdata = tpm2.importkey(
-            pobj_handle, pobj['objauth'], objauth, privkey=privkey, alg=alg)
+            pobj_handle, pobj['objauth'], objauth, privkey=privkey, alg=alg, passin=passin, objattrs=objattrs)
+
+        # We have no way of knowing the keylength of an hmac key
+        if alg and alg.startswith('hmac') or alg == 'keyedhash':
+            self._override_keylen = os.path.getsize(privkey)
 
         return (tertiarypriv, tertiarypub, tertiarypubdata)
 
@@ -247,7 +278,7 @@ class AddKeyCommand(NewKeyCommandBase):
             required=True)
 
     # Creates a new key
-    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, privkey, passin, d):
 
         pobj_handle = get_pobject(pobj, tpm2, hierarchyauth, d)
 
@@ -584,7 +615,8 @@ class LinkCommand(NewKeyCommandBase):
             sys.exit('The primary object (id: {:d}) is persistent and'
                 ' the TSS Engine key does not have a persistent parent,'
                 ' got: 0x{:x}'.format(pid, phandle))
-        elif is_transient and phandle != 0:
+        elif is_transient and not (phandle == tpm2.TPM2_RH_OWNER or phandle == 0):
+            # tpm2-tss-engine < 1.1.0 used a phandle of 0 instead of tpm2.TPM2_RH_OWNER
             sys.exit('The primary object (id: {:d}) is transient and'
                 ' the TSS Engine key has a parent handle,'
                 ' got: 0x{:x}'.format(pid, phandle))
@@ -641,7 +673,7 @@ class LinkCommand(NewKeyCommandBase):
         return (tertiarypriv, tertiarypub, tertiarypubdata)
 
     # Links a new key
-    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, d):
+    def new_key_create(self, pobj, objauth, hierarchyauth, tpm2, alg, keypaths, passin, d):
 
         if keypaths is None:
             sys.exit("Keypath must be specified")
