@@ -8,6 +8,10 @@ import uuid
 
 from subprocess import Popen, PIPE
 
+from cryptography.hazmat.primitives import serialization
+from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
+from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
+
 from .utils import str2bytes
 
 class Tpm2(object):
@@ -19,8 +23,10 @@ class Tpm2(object):
         "|fixedparent|sensitivedataorigin"
 
     ALGS = [
-        'rsa1024', 'rsa2048', 'aes128', 'aes256', 'ecc224', 'ecc256',
-        'ecc384', 'ecc521'
+        'rsa1024', 'rsa2048', 'rsa3072', 'rsa4096',
+        'aes128', 'aes256',
+        'ecc224', 'ecc256', 'ecc384', 'ecc521',
+        'hmac:sha1', 'hmac:sha256', 'hmac:sha384', 'hmac:sha512'
     ]
 
     TEMPLATES = {
@@ -96,7 +102,7 @@ class Tpm2(object):
         p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
         stdout, stderr = p.communicate()
         if (p.wait()):
-            raise RuntimeError("Could not execute tpm2_evictcontrol: %s",
+            raise RuntimeError("Could not execute tpm2_evictcontrol: %s" %
                                stderr)
         return tr_file
 
@@ -112,7 +118,7 @@ class Tpm2(object):
         p = Popen(cmd, stdout=PIPE, stderr=PIPE, env=os.environ)
         stdout, stderr = p.communicate()
         if (p.wait()):
-            raise RuntimeError("Could not execute tpm2_readpublic: %s",
+            raise RuntimeError("Could not execute tpm2_readpublic: %s" %
                                stderr)
         return (stdout, tr_file if get_tr_file else None)
 
@@ -146,7 +152,7 @@ class Tpm2(object):
         _, stderr = p.communicate()
         rc = p.wait()
         if rc:
-            raise RuntimeError("Could not execute tpm2_load: %s", stderr)
+            raise RuntimeError("Could not execute tpm2_load: %s" % stderr)
         return ctx
 
     def unseal(self, ctx, auth):
@@ -157,7 +163,7 @@ class Tpm2(object):
         stdout, stderr = p.communicate()
         rc = p.wait()
         if rc:
-            raise RuntimeError("Could not execute tpm2_unseal: %s", stderr)
+            raise RuntimeError("Could not execute tpm2_unseal: %s" % stderr)
         return stdout
 
     def _encryptdecrypt(self, ctx, auth, data, decrypt=False):
@@ -171,7 +177,7 @@ class Tpm2(object):
         stdout, stderr = p.communicate(input=data)
         rc = p.wait()
         if rc:
-            raise RuntimeError("Could not execute tpm2_encryptdecrypt: %s",
+            raise RuntimeError("Could not execute tpm2_encryptdecrypt: %s" %
                                stderr)
         return stdout
 
@@ -189,8 +195,10 @@ class Tpm2(object):
                seal=None,
                alg=None):
         # tpm2_create -Q -C context.out -g $gAlg -G $GAlg -u key.pub -r key.priv
-        _, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
-        _, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
+        privfd, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
+        os.close(privfd)
+        pubfd, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
+        os.close(pubfd)
 
         cmd = ['tpm2_create', '-C', str(phandle), '-u', pub, '-r', priv]
 
@@ -228,7 +236,7 @@ class Tpm2(object):
         stdout, stderr = p.communicate()
         rc = p.wait()
         if rc:
-            raise RuntimeError("Could not execute tpm2_getcap: %s", stderr)
+            raise RuntimeError("Could not execute tpm2_getcap: %s" % stderr)
         return stdout
 
     def importkey(self,
@@ -238,18 +246,74 @@ class Tpm2(object):
                   privkey,
                   objattrs=None,
                   seal=None,
-                  alg=None):
+                  alg=None,
+                  passin=None):
 
         if privkey and len(privkey) > 0:
             exists = os.path.isfile(privkey)
             if not exists:
-                raise RuntimeError("File '%s' path is invalid or is missing",
+                raise RuntimeError("File '%s' path is invalid or is missing" %
                                    privkey)
         else:
             sys.exit("Invalid file path")
 
-        _, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
-        _, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
+        privfd, priv = mkstemp(prefix='', suffix='.priv', dir=self._tmp)
+        os.close(privfd)
+        pubfd, pub = mkstemp(prefix='', suffix='.pub', dir=self._tmp)
+        os.close(pubfd)
+
+        # If the key is an OpenSSH key, convert it to PEM format
+        pem_priv_name = None
+        with open(privkey, "rb") as f:
+            privey_data = f.read()
+        if privey_data.startswith(b'-----BEGIN OPENSSH PRIVATE KEY-----'):
+            if passin:
+                # Parse passin to extract the password
+                if passin.startswith('env:'):
+                    password_bytes = os.getenvb(passin[4:].encode())
+                elif passin.startswith('file:'):
+                    with open(passin[5:], 'rb') as f:
+                        password_bytes = f.read()
+                elif passin.startswith('pass:'):
+                    password_bytes = passin[5:].encode()
+                else:
+                    raise NotImplementedError("Unsupported OpenSSL password input {} to read OpenSSH key".format(
+                        repr(passin)))
+                enc_alg = serialization.BestAvailableEncryption(password_bytes)
+            else:
+                password_bytes = None
+                enc_alg = serialization.NoEncryption()
+
+            ssh_key = serialization.load_ssh_private_key(privey_data, password=password_bytes)
+
+            if alg is None:
+                # Find the algorithm
+                if isinstance(ssh_key, EllipticCurvePrivateKey):
+                    alg = 'ecc'
+                elif isinstance(ssh_key, RSAPrivateKey):
+                    alg = 'rsa'
+                else:
+                    raise NotImplementedError("Unsupported SSH key type {}".format(type(ssh_key)))
+
+            pem_key = ssh_key.private_bytes(
+                encoding=serialization.Encoding.PEM,
+                format=serialization.PrivateFormat.TraditionalOpenSSL,
+                encryption_algorithm=enc_alg)
+
+            pem_priv_fd, pem_priv_name = mkstemp(prefix='', suffix='.privpem', dir=self._tmp)
+            os.write(pem_priv_fd, pem_key)
+            os.close(pem_priv_fd)
+            privkey = pem_priv_name
+        elif alg is None:
+            # Guess the key algorithm from the PEM header
+            if privey_data.startswith(b'-----BEGIN EC PARAMETERS-----'):
+                alg = 'ecc'
+            elif privey_data.startswith(b'-----BEGIN EC PRIVATE KEY-----'):
+                alg = 'ecc'
+            elif privey_data.startswith(b'-----BEGIN RSA PRIVATE KEY-----'):
+                alg = 'rsa'
+            else:
+                raise RuntimeError("Unable to detect key type, use --algorithm to specify it")
 
         parent_path = str(phandle)
         cmd = [
@@ -272,15 +336,20 @@ class Tpm2(object):
         if alg != None:
             cmd.extend(['-G', alg])
 
+        if passin is not None:
+            cmd.extend(['--passin', passin])
+
         p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE, env=os.environ)
         stdout, stderr = p.communicate(input=seal)
         rc = p.wait()
+        if pem_priv_name is not None:
+            os.remove(pem_priv_name)
         if (rc != 0):
             os.remove(pub)
             os.remove(priv)
             print("command: %s" % str(" ".join(cmd)))
             raise RuntimeError("Could not execute tpm2_import: %s" %
-                               str(stderr))
+                               stderr)
 
         return priv, pub, stdout
 
@@ -304,7 +373,7 @@ class Tpm2(object):
         _, stderr = p.communicate()
         rc = p.wait()
         if rc:
-            raise RuntimeError("Could not execute tpm2_changeauth: %s", stderr)
+            raise RuntimeError("Could not execute tpm2_changeauth: %s" % stderr)
 
         return newpriv
 
